import json
import os
import typing

from agentlys.base import AgentlysBase
from agentlys.model import Message, MessagePart
from agentlys.providers.base_provider import BaseProvider
from agentlys.providers.utils import FunctionCallParsingError

AGENTLYS_HOST = os.getenv("AGENTLYS_HOST")


def from_openai_object(
    role: str,
    content: str,
    function_call: typing.Optional[dict] = None,
    id: typing.Optional[str] = None,
):
    # We need to unquote the arguments
    function_call_dict = None
    if function_call:
        try:
            function_call_dict = {
                "name": function_call.name,
                "arguments": json.loads(function_call.arguments),
            }
        except json.decoder.JSONDecodeError:
            raise FunctionCallParsingError(id, function_call)

    return Message(role=role, content=content, function_call=function_call_dict, id=id)


def parts_to_openai_dict(part: MessagePart) -> dict:
    if part.type == "text":
        return {
            "type": "text",
            "text": part.content,
        }
    elif part.type == "image":
        return {
            "type": "image_url",
            "image_url": {
                "url": f"data:{part.image.format};base64,{part.image.to_base64()}"
            },
        }
    elif part.type == "function_call":
        return {
            "name": part.function_call["name"],
            "arguments": json.dumps(part.function_call["arguments"]),
        }
    elif part.type == "function_result":
        return {
            "type": "text",
            "text": part.content,
        }
    elif part.type == "function_result_image":
        return {
            "type": "image_url",
            "image_url": {
                "url": f"data:{part.image.format};base64,{part.image.to_base64()}"
            },
        }

    raise ValueError(f"Unknown part type: {part.type}")


def message_to_openai_dict(message: Message) -> dict:
    res = {
        "role": message.role,
        "content": [],
    }

    for part in message.parts:
        if part.type == "function_call" and message.role == "assistant":
            res["function_call"] = parts_to_openai_dict(part)
        elif part.type == "function_call" and message.role == "user":
            # Workaround: If the user is triggering a function, we add it's name and arguments to the content
            res["content"] = (
                part.function_call["name"]
                + ":"
                + json.dumps(part.function_call["arguments"])
            )
        else:
            res["content"].append(parts_to_openai_dict(part))

    if message.name:
        res["name"] = message.name

    # Workaround
    # Image URLs are only allowed for messages with role 'user'
    # If the message contains an image, we change it's role to 'user'
    if message.role == "function" and any(
        part.type == "function_result_image" for part in message.parts
    ):
        res["role"] = "user"
    return res


class OpenAIProviderFunctionLegacy(BaseProvider):
    def __init__(self, chat: AgentlysBase, model: str, base_url: str = None):
        from openai import OpenAI

        self.chat = chat
        self.model = model
        # Possibly set up openai.api_key, base_url, etc.
        self.client = OpenAI(
            base_url=(
                f"{AGENTLYS_HOST}/v1" if AGENTLYS_HOST else "https://api.openai.com/v1"
            ),
        )

    async def fetch_async(self, **kwargs) -> Message:
        messages = self.prepare_messages(transform_function=message_to_openai_dict)
        # Add instruction as the first message

        system_messages = []
        if self.chat.instruction:
            system_messages.append(
                Message(
                    role="system",
                    content=self.chat.instruction,
                )
            )
        if self.chat.last_tools_states:
            system_messages.append(
                Message(
                    role="system",
                    content=self.chat.last_tools_states,
                )
            )
        messages = [message_to_openai_dict(sm) for sm in system_messages] + messages

        if self.chat.functions_schema:
            res = self.client.chat.completions.create(
                model=self.model,
                messages=messages,
                functions=self.chat.functions_schema,
                **kwargs,
            )
        else:
            res = self.client.chat.completions.create(
                model=self.model, messages=messages, **kwargs
            )

        message = res.choices[0].message
        return from_openai_object(
            role=message.role,
            content=message.content,
            function_call=message.function_call,
            id=res.id,  # We use the response id as the message id
        )



"""
This is a demo of how to use the openai_function_shim provider,
which is a hacky way to use function calling with o1 OpenAI models.

It is not recommended for production use, but it can be useful for
testing and experimentation.
"""

import pytest
from agentlys import Agentlys, APIProvider


def multiply(a: int, b: int) -> int:
    return a * b


@pytest.mark.vcr
def test_openai_function_shim():
    # The only model thing you need to do is set the provider to "openai_function_shim"
    agent = Agentlys(provider=APIProvider.OPENAI_FUNCTION_SHIM, model="o1-preview")
    agent.add_function(multiply)

    text = "What is 343354 * 13243343214"
    for message in agent.run_conversation(text):
        print(message.to_markdown())


if __name__ == "__main__":
    test_openai_function_shim()



import typing
from abc import ABC, abstractmethod
from enum import Enum

from agentlys.model import Message, MessagePart
from agentlys.utils import get_event_loop_or_create


class APIProvider(Enum):
    OPENAI = "openai"
    OPENAI_FUNCTION_LEGACY = "openai_function_legacy"
    OPENAI_FUNCTION_SHIM = "openai_function_shim"
    ANTHROPIC = "anthropic"
    DEFAULT = "default"


class BaseProvider(ABC):
    def prepare_messages(
        self,
        transform_function: typing.Callable,
        transform_list_function: typing.Callable = lambda x: x,
    ) -> list[dict]:
        """Prepare messages for API requests using a transformation function."""
        first_message = self.chat.messages[0]

        # Add combined context to the first message if it exists
        if self.chat.context:
            if isinstance(first_message.content, str):
                first_message.parts[0].content = (
                    self.chat.context + "\n" + first_message.parts[0].content
                )
            elif isinstance(first_message.content, list):
                first_message.content = [
                    MessagePart(type="text", content=self.chat.context),
                    *first_message.content,
                ]

        messages = self.chat.examples + [first_message] + self.chat.messages[1:]
        messages = transform_list_function(messages)
        return [transform_function(m) for m in messages]

    @abstractmethod
    async def fetch_async(self, **kwargs) -> Message:
        """
        Async version of fetch method.
        Given a chat context, returns a single new Message from the LLM.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def fetch(self, **kwargs) -> Message:
        """
        Given a chat context, returns a single new Message from the LLM.
        """
        # For backward compatibility, use run_until_complete to execute the async method
        loop = get_event_loop_or_create()
        return loop.run_until_complete(self.fetch_async(**kwargs))

    async def fetch_stream_async(self, **kwargs) -> typing.AsyncGenerator[dict, None]:
        """
        Async streaming version of fetch method.
        Yields chunks as they arrive from the LLM.

        Yields:
            - {"type": "text", "content": str} - text chunks as they arrive
            - {"type": "message", "message": Message} - final complete message

        Note: Subclasses should override this method to provide streaming support.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support streaming. "
            "Please implement fetch_stream_async or use a provider that supports streaming."
        )
        # This yield is needed to make this a generator function
        yield  # type: ignore  # pragma: no cover



import unittest
from unittest.mock import patch

from agentlys import Agentlys, Message
from agentlys.providers.openai import OpenAIProvider


class TestAgentlys(unittest.TestCase):
    def test_agentlys_initialization(self):
        agent = Agentlys(instruction="Test instruction", provider="openai")
        self.assertEqual(agent.instruction, "Test instruction")
        self.assertEqual(agent.provider.__class__, OpenAIProvider)
        # Instead of checking for a specific model, we'll just check if it's a string
        self.assertIsInstance(agent.model, str)
        self.assertTrue(len(agent.model) > 0)  # Ensure the model is not an empty string

    def test_agentlys_invalid_provider(self):
        with self.assertRaises(ValueError):
            Agentlys(provider="invalid_provider")

    def test_add_function(self):
        agent = Agentlys()

        def test_function(arg1: str, arg2: int) -> str:
            return f"Received {arg1} and {arg2}"

        agent.add_function(test_function)
        self.assertEqual(len(agent.functions_schema), 1)
        self.assertIn("test_function", agent.functions)

    @patch.object(OpenAIProvider, "fetch_async")
    def test_ask(self, mock_fetch_openai):
        mock_fetch_openai.return_value = Message(
            role="assistant", content="Test response"
        )
        agent = Agentlys(provider="openai")

        response = agent.ask("Test question")
        self.assertEqual(response.role, "assistant")
        self.assertEqual(response.content, "Test response")
        self.assertEqual(len(agent.messages), 2)

    @patch.object(OpenAIProvider, "fetch_async")
    def test_run_conversation(self, mock_fetch_openai):
        mock_fetch_openai.return_value = Message(
            role="assistant", content="Final response"
        )
        agent = Agentlys(provider="openai")

        responses = list(agent.run_conversation("Test question"))
        self.assertEqual(len(responses), 2)
        self.assertEqual(responses[0].role, "user")
        self.assertEqual(responses[0].content, "Test question")
        self.assertEqual(responses[1].role, "assistant")
        self.assertEqual(responses[1].content, "Final response")


if __name__ == "__main__":
    unittest.main()



import json
import os

from agentlys.base import AgentlysBase
from agentlys.model import Message
from agentlys.providers.openai import (
    OpenAIProvider,
    parts_to_openai_dict,
)

AGENTLYS_HOST = os.getenv("AGENTLYS_HOST")


def message_to_openai_dict(message: Message) -> dict:
    if message.role == "function":
        res = {
            "role": "tool",
            "tool_call_id": message.function_call_id,
            "content": message.content,
            "name": message.name,
        }
    elif message.role == "system":
        res = {"role": message.role, "content": message.content}
    else:
        res = {"role": message.role, "content": []}
        for part in message.parts:
            if part.type == "function_call" and message.role == "assistant":
                res["tool_calls"] = [
                    {
                        "id": part.function_call_id,
                        "type": "function",
                        "function": parts_to_openai_dict(part),
                    }
                ]
            elif part.type == "function_call" and message.role == "user":
                # Workaround: If the user is triggering a function, we add it's name and arguments to the content
                res["content"] = (
                    part.function_call["name"]
                    + ":"
                    + json.dumps(part.function_call["arguments"])
                )
            else:
                res["content"].append(parts_to_openai_dict(part))

        if "content" in res and len(res["content"]) == 0:
            res["content"] = None

    return res


class DefaultProvider(OpenAIProvider):
    """Default provider is OpenAI with small changes
    - System messages are string only
    - Function call content is a string
    """

    def __init__(self, chat: AgentlysBase, model: str, base_url: str = None):
        from openai import OpenAI

        self.chat = chat
        self.model = model
        self.client = OpenAI(base_url=AGENTLYS_HOST)



import os

import anthropic
from agentlys.base import AgentlysBase
from agentlys.model import Message, MessagePart
from agentlys.providers.base_provider import BaseProvider
from agentlys.providers.utils import add_empty_function_result

AGENTLYS_HOST = os.getenv("AGENTLYS_HOST")


def part_to_anthropic_dict(part: MessagePart) -> dict:
    if part.type == "text":
        return {
            "type": "text",
            "text": part.content,
        }
    elif part.type == "image":
        return {
            "type": "image",
            "source": {
                "type": "base64",
                "media_type": part.image.format,
                "data": part.image.to_base64(),
            },
        }
    elif part.type == "function_call":
        return {
            "type": "tool_use",
            "id": part.function_call_id,
            "name": part.function_call["name"],
            "input": part.function_call["arguments"],
        }
    elif part.type == "function_result_image":
        return {
            "type": "tool_result",
            "content": [
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": part.image.format,
                        "data": part.image.to_base64(),
                    },
                }
            ],
            "tool_use_id": part.function_call_id,
        }
    elif part.type == "function_result":
        return {
            "type": "tool_result",
            "tool_use_id": part.function_call_id,
            "content": part.content,
        }
    elif part.type == "thinking":
        if part.is_redacted:
            return {
                "type": "redacted_thinking",
                "data": part.thinking_signature,
            }
        return {
            "type": "thinking",
            "thinking": part.thinking,
            "signature": part.thinking_signature,
        }
    raise ValueError(f"Unknown part type: {part.type}")


def message_to_anthropic_dict(message: Message) -> dict:
    res = {
        "role": message.role if message.role in ["user", "assistant"] else "user",
        "content": [],
    }

    for part in message.parts:
        res["content"].append(part_to_anthropic_dict(part))

    return res


DEFAULT_MAX_TOKENS = int(os.getenv("ANTHROPIC_MAX_TOKENS", "10000"))


class AnthropicProvider(BaseProvider):
    def __init__(self, chat: AgentlysBase, model: str, max_tokens: int | None = None):
        self.model = model
        self.client = anthropic.AsyncAnthropic(
            base_url=AGENTLYS_HOST if AGENTLYS_HOST else "https://api.anthropic.com",
        )
        self.chat = chat
        self.max_tokens = DEFAULT_MAX_TOKENS if max_tokens is None else max_tokens

    def _prepare_request_params(self, **kwargs):
        """Prepare messages, tools, and kwargs for Anthropic API request."""
        messages = self.prepare_messages(
            transform_function=lambda m: message_to_anthropic_dict(m),
            transform_list_function=add_empty_function_result,
        )

        if self.chat.instruction:
            system = self.chat.instruction
        else:
            system = None

        def merge_messages(messages):
            """
            When two messages are in the same role, we merge the following message into the previous.
            {
                "role": "user",
                "content": [
                    {
                    "type": "tool_result",
                    "tool_use_id": "example_19",
                    "content": ""
                    }
                ]
            },
            {
                "role": "user",
                "content": "Plot distribution of stations per city"
            }
            """
            merged_messages = []
            for message in messages:
                if merged_messages and merged_messages[-1]["role"] == message["role"]:
                    if isinstance(merged_messages[-1]["content"], str):
                        merged_messages[-1]["content"].append(
                            {
                                "type": "text",
                                "text": merged_messages[-1]["content"],
                            }
                        )
                    elif isinstance(merged_messages[-1]["content"], list):
                        merged_messages[-1]["content"].extend(message["content"])
                    else:
                        raise ValueError(
                            f"Invalid content type: {type(merged_messages[-1]['content'])}"
                        )
                else:
                    merged_messages.append(message)
            return merged_messages

        messages = merge_messages(messages)

        # Need to map field "parameters" to "input_schema"
        tools = [
            {
                "name": s["name"],
                "description": s["description"],
                "input_schema": s["parameters"],
            }
            for s in self.chat.functions_schema
        ]
        # Add description to the function is their description is empty
        for tool in tools:
            if not tool["description"]:
                tool["description"] = "No description provided"

        # === Add cache_controls ===
        # Messages: Find the last message with an index multiple of 10
        last_message_index = next(
            (i for i in reversed(range(len(messages))) if i % 10 == 0),
            None,
        )

        if last_message_index is not None:
            # Only try to modify the cache control if there are messages and content
            if (
                messages
                and isinstance(messages[last_message_index]["content"], list)
                and len(messages[last_message_index]["content"]) > 0
                and isinstance(messages[last_message_index]["content"][-1], dict)
            ):
                messages[last_message_index]["content"][-1]["cache_control"] = {
                    "type": "ephemeral"
                }
            elif isinstance(messages[last_message_index]["content"], str):
                messages[last_message_index]["content"] = [
                    {
                        "type": "text",
                        "text": messages[last_message_index]["content"],
                        "cache_control": {"type": "ephemeral"},
                    }
                ]
        # Tools: Add cache_control to the last tool function
        if tools:
            tools[-1]["cache_control"] = {"type": "ephemeral"}

        # System: add cache_control to the system message
        system_messages = []
        if system is not None:
            system_messages.append(
                {
                    "type": "text",
                    "text": system,
                    "cache_control": {"type": "ephemeral"},
                }
            )

        if self.chat.last_tools_states:
            # add to system message
            system_messages.append(
                {
                    "type": "text",
                    "text": self.chat.last_tools_states,
                }
            )

        # === End of cache_control ===

        if system_messages:
            kwargs["system"] = system_messages

        if self.chat.use_tools_only and "tool_choice" not in kwargs:
            kwargs["tool_choice"] = {"type": "any"}

        # Add thinking config if set at class level and not already in kwargs
        if getattr(self.chat, "thinking", None) and "thinking" not in kwargs:
            kwargs["thinking"] = self.chat.thinking

        return messages, tools, kwargs

    async def fetch_async(self, **kwargs):
        messages, tools, kwargs = self._prepare_request_params(**kwargs)

        res = await self.client.messages.create(
            model=self.model,
            messages=messages,
            tools=tools,
            max_tokens=self.max_tokens,
            **kwargs,
        )
        res_dict = res.to_dict()
        return Message.from_anthropic_dict(
            role=res_dict["role"],
            content=res_dict["content"],
        )

    async def fetch_stream_async(self, **kwargs):
        """Stream response tokens from Anthropic.

        Yields text chunks as they arrive. Returns the final Message
        (with potential tool calls) after streaming completes.
        """
        messages, tools, kwargs = self._prepare_request_params(**kwargs)

        async with self.client.messages.stream(
            model=self.model,
            messages=messages,
            tools=tools if tools else anthropic.NOT_GIVEN,
            max_tokens=self.max_tokens,
            **kwargs,
        ) as stream:
            async for event in stream:
                if event.type == "content_block_delta":
                    if event.delta.type == "thinking_delta":
                        yield {"type": "thinking", "content": event.delta.thinking}
                    elif event.delta.type == "text_delta":
                        yield {"type": "text", "content": event.delta.text}

            # Get final message for tool handling
            response = await stream.get_final_message()
            res_dict = response.to_dict()
            final_message = Message.from_anthropic_dict(
                role=res_dict["role"],
                content=res_dict["content"],
            )
            yield {"type": "message", "message": final_message}
